"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Migration = void 0;
const fs = require("fs");
const path = require("path");
const util_1 = require("util");
try {
    require('dotenv').config();
}
catch (err) { }
class Migration {
    constructor(config) {
        var _a, _b, _c;
        this.scripts = [];
        this.maxVersion = 0;
        this.isInit = false;
        this.strictOrder = !!process.env.MIGRATIONS_STRICT_ORDER
            ? ((_a = process.env.MIGRATIONS_STRICT_ORDER) === null || _a === void 0 ? void 0 : _a.toLowerCase()) === 'true'
            : ((_b = process.env.NODE_ENV) === null || _b === void 0 ? void 0 : _b.toLowerCase()) === 'production';
        this.numOrder = ((_c = process.env.MIGRATIONS_NUMERIC_ORDER) === null || _c === void 0 ? void 0 : _c.toLowerCase()) === 'true';
        this.config = config;
        this.isInit = false;
        this.strictOrder =
            config.strictOrder !== undefined ? config.strictOrder : this.strictOrder;
        this.numOrder =
            config.numOrder !== undefined ? config.numOrder : this.numOrder;
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.initMigrationTable();
            yield this.loadScripts(this.config.dir || 'migrations');
            this.isInit = true;
        });
    }
    up(steps) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.isInit) {
                throw new Error('Migration class not initialized! Run initialize function first!');
            }
            this.writeLog('Starting upgrade migration!');
            const infinite = !steps || steps < 0;
            let countStep = 0;
            while (infinite || countStep < steps) {
                const script = yield this.getNextUpgradeScript();
                if (!script) {
                    const { version } = yield this.getLastVersion();
                    if (version != this.maxVersion) {
                        throw new Error('Next upgrade script not found!');
                    }
                    else {
                        this.writeLog('Upgrade complete!');
                        break;
                    }
                }
                else if (script.upgrade) {
                    this.writeLog(`Step: ${countStep + 1}`);
                    this.writeLog(`Upgrading to version ${script.version}...`);
                    yield script.upgrade(this.query.bind(this));
                    yield this.saveVersion(script);
                    this.writeLog(`DB upgraded to version : ${script.version}`);
                }
                countStep++;
            }
        });
    }
    down(steps = 1) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.isInit) {
                throw new Error('Migration class not initialized! Run initialize function first!');
            }
            this.writeLog('Starting downgrade migration!');
            const infinite = !steps || steps < 0;
            let countStep = 0;
            while (infinite || countStep < steps) {
                const script = yield this.getNextDowngradeScript();
                if (!script) {
                    const { version } = yield this.getLastVersion();
                    if (version) {
                        throw new Error('Next downgrade script not found!');
                    }
                    else {
                        this.writeLog('Downgrade complete!');
                        break;
                    }
                }
                else if (script.downgrade) {
                    this.writeLog(`Step: ${countStep + 1}`);
                    this.writeLog(`Downgrading to version ${script.version - 1}...`);
                    yield script.downgrade(this.query.bind(this));
                    yield this.deleteVersion(script);
                    this.writeLog(`DB downgraded to version : ${script.version - 1}`);
                }
                countStep++;
            }
        });
    }
    reset() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.down(-1);
            yield this.up();
        });
    }
    getLastVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.query(`SELECT version, fileName
      FROM ${this.config.tableName}
      ORDER BY version DESC
      LIMIT 1`, null, true).then((val) => !!val && val.length
                ? { version: val[0].version, fileName: val[0].fileName }
                : { version: 0, fileName: null });
        });
    }
    getLoadedScripts(sortDirection = 'desc') {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.query(`SELECT version, fileName
      FROM ${this.config.tableName}
      ORDER BY version ${sortDirection}`, null, true);
        });
    }
    destroy() {
        return __awaiter(this, void 0, void 0, function* () {
            yield new Promise((resolve, reject) => {
                this.config.conn.end((e) => {
                    this.writeLog(`Migrations: DB connection terminated: ${(e === null || e === void 0 ? void 0 : e.message) || 'CLOSED'}`);
                    resolve();
                });
            });
        });
    }
    loadScripts(dirPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const files = yield (0, util_1.promisify)(fs.readdir)(dirPath);
            let ver = 1;
            const loaded = yield this.getLoadedScripts('asc');
            const notLoaded = files.filter((file) => !loaded.find((loadedScripts) => loadedScripts.fileName === file));
            const sortedNotLoadedScripts = notLoaded.sort(this.sortFiles);
            const fileArr = [
                ...loaded.map((script) => script.fileName),
                ...sortedNotLoadedScripts,
            ];
            this.writeLog(`Found migration scripts: ${fileArr.join(', ')}`);
            let previousFilePrefix = null;
            fileArr.forEach((file) => {
                let script;
                try {
                    script = module.parent.parent.require(path.resolve(process.cwd(), dirPath, file));
                }
                catch (e) {
                    this.writeLog(`Unable to load script from ${file}! (${e})`);
                    if (/(\.js|\.ts|\.jsx|\.jsm)$/.test(file)) {
                        throw e;
                    }
                }
                const filePrefix = file.match(/^(\d*)-/)
                    ? file.match(/^(\d*)-/)[1]
                    : null;
                if (filePrefix &&
                    previousFilePrefix &&
                    ((this.strictOrder && +filePrefix < +previousFilePrefix) ||
                        (this.numOrder && +previousFilePrefix + 1 !== +filePrefix))) {
                    throw new Error('Check your migration scripts! You are trying to load a script with a lower timestamp or version index than the last loaded script!');
                }
                const isValid = !!script &&
                    typeof script.upgrade !== 'undefined' &&
                    typeof script.downgrade !== 'undefined';
                if (isValid) {
                    script.version = ver;
                    script.fileName = file;
                    this.scripts.push(script);
                    this.maxVersion = ver;
                    ver++;
                    previousFilePrefix = filePrefix || previousFilePrefix;
                }
            });
        });
    }
    getNextUpgradeScript() {
        return __awaiter(this, void 0, void 0, function* () {
            const { version } = yield this.getLastVersion();
            for (const script of this.scripts) {
                if (script.version === version + 1) {
                    return script;
                }
            }
            return null;
        });
    }
    getNextDowngradeScript() {
        return __awaiter(this, void 0, void 0, function* () {
            const { version, fileName } = yield this.getLastVersion();
            for (const script of this.scripts) {
                if (script.version === version) {
                    if (script.fileName !== fileName) {
                        throw new Error(`Script version and filename mismatch! version: ${version} -> file system: ${script.fileName} | database: ${fileName}`);
                    }
                    return script;
                }
            }
            return null;
        });
    }
    initMigrationTable() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.query(`
        CREATE TABLE IF NOT EXISTS ${this.config.tableName} (
          version INT NOT NULL,
          fileName VARCHAR(256),
          date DATETIME DEFAULT CURRENT_TIMESTAMP,
          PRIMARY KEY (version)
        )
      `, null, true);
            }
            catch (err) {
                this.writeLog('Failed to create table for migrations.');
                throw new Error(err);
            }
        });
    }
    saveVersion(script) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.query(`INSERT INTO ${this.config.tableName} (version, fileName)
       VALUES (${script.version}, '${script.fileName}')
      `, null, true);
        });
    }
    deleteVersion(script) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.query(`DELETE FROM ${this.config.tableName}
       WHERE version = ${script.version}
      `, null, true);
        });
    }
    query(query, values, silent = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const q = (0, util_1.promisify)(this.config.conn.query).bind(this.config.conn);
            this.writeLog(query, silent);
            return yield q(query, values);
        });
    }
    sortFiles(a, b) {
        const verA = a.match(/^(\d*)-/)[1];
        const verB = b.match(/^(\d*)-/)[1];
        if (!parseInt(verA) || !parseInt(verB)) {
            return 0;
        }
        if (parseInt(verA) > parseInt(verB)) {
            return 1;
        }
        if (parseInt(verA) < parseInt(verB)) {
            return -1;
        }
        return 0;
    }
    writeLog(text, forceSilent) {
        if (!this.config.silent && !forceSilent) {
            console.log(text);
        }
    }
}
exports.Migration = Migration;
//# sourceMappingURL=migrations.js.map